// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import lineIntersect from '@turf/line-intersect';
import { lineString as turfLineString } from '@turf/helpers';
import { getPickedEditHandle } from "./utils.js";
import { GeoJsonEditMode } from "./geojson-edit-mode.js";
export class DrawPolygonMode extends GeoJsonEditMode {
    createTentativeFeature(props) {
        const { lastPointerMoveEvent } = props;
        const clickSequence = this.getClickSequence();
        const lastCoords = lastPointerMoveEvent ? [lastPointerMoveEvent.mapCoords] : [];
        let tentativeFeature;
        if (clickSequence.length === 1 || clickSequence.length === 2) {
            tentativeFeature = {
                type: 'Feature',
                properties: {
                    guideType: 'tentative'
                },
                geometry: {
                    type: 'LineString',
                    coordinates: [...clickSequence, ...lastCoords]
                }
            };
        }
        else if (clickSequence.length > 2) {
            tentativeFeature = {
                type: 'Feature',
                properties: {
                    guideType: 'tentative'
                },
                geometry: {
                    type: 'Polygon',
                    coordinates: [[...clickSequence, ...lastCoords, clickSequence[0]]]
                }
            };
        }
        return tentativeFeature;
    }
    getGuides(props) {
        const clickSequence = this.getClickSequence();
        const guides = {
            type: 'FeatureCollection',
            features: []
        };
        const tentativeFeature = this.createTentativeFeature(props);
        if (tentativeFeature) {
            guides.features.push(tentativeFeature);
        }
        const editHandles = clickSequence.map((clickedCoord, index) => ({
            type: 'Feature',
            properties: {
                guideType: 'editHandle',
                editHandleType: 'existing',
                featureIndex: -1,
                positionIndexes: [index]
            },
            geometry: {
                type: 'Point',
                coordinates: clickedCoord
            }
        }));
        guides.features.push(...editHandles);
        return guides;
    }
    // eslint-disable-next-line complexity
    handleClick(event, props) {
        const { picks } = event;
        const clickedEditHandle = getPickedEditHandle(picks);
        const clickSequence = this.getClickSequence();
        let overlappingLines = false;
        if (clickSequence.length > 2 && props.modeConfig && props.modeConfig.preventOverlappingLines) {
            const currentLine = turfLineString([
                clickSequence[clickSequence.length - 1],
                event.mapCoords
            ]);
            const otherLines = turfLineString([...clickSequence.slice(0, clickSequence.length - 1)]);
            const intersectingPoints = lineIntersect(currentLine, otherLines);
            if (intersectingPoints.features.length > 0) {
                overlappingLines = true;
            }
        }
        let positionAdded = false;
        if (!clickedEditHandle && !overlappingLines) {
            // Don't add another point right next to an existing one
            this.addClickSequence(event);
            positionAdded = true;
        }
        if (clickSequence.length > 2 &&
            clickedEditHandle &&
            Array.isArray(clickedEditHandle.properties.positionIndexes) &&
            (clickedEditHandle.properties.positionIndexes[0] === 0 ||
                clickedEditHandle.properties.positionIndexes[0] === clickSequence.length - 1)) {
            // They clicked the first or last point (or double-clicked), so complete the polygon
            // Remove the hovered position
            const polygonToAdd = {
                type: 'Polygon',
                coordinates: [[...clickSequence, clickSequence[0]]]
            };
            this.resetClickSequence();
            const editAction = this.getAddFeatureOrBooleanPolygonAction(polygonToAdd, props);
            if (editAction) {
                props.onEdit(editAction);
            }
        }
        else if (positionAdded) {
            // new tentative point
            props.onEdit({
                // data is the same
                updatedData: props.data,
                editType: 'addTentativePosition',
                editContext: {
                    position: event.mapCoords
                }
            });
        }
    }
    handleKeyUp(event, props) {
        if (event.key === 'Enter') {
            const clickSequence = this.getClickSequence();
            if (clickSequence.length > 2) {
                const polygonToAdd = {
                    type: 'Polygon',
                    coordinates: [[...clickSequence, clickSequence[0]]]
                };
                this.resetClickSequence();
                const editAction = this.getAddFeatureOrBooleanPolygonAction(polygonToAdd, props);
                if (editAction) {
                    props.onEdit(editAction);
                }
            }
        }
        else if (event.key === 'Escape') {
            this.resetClickSequence();
            props.onEdit({
                // Because the new drawing feature is dropped, so the data will keep as the same.
                updatedData: props.data,
                editType: 'cancelFeature',
                editContext: {}
            });
        }
    }
    handlePointerMove(event, props) {
        props.onUpdateCursor('cell');
        super.handlePointerMove(event, props);
    }
}
