// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import turfUnion from '@turf/union';
import turfDifference from '@turf/difference';
import turfIntersect from '@turf/intersect';
import rewind from '@turf/rewind';
import { getPickedEditHandles, getNonGuidePicks } from "./utils.js";
import { ImmutableFeatureCollection } from "./immutable-feature-collection.js";
const DEFAULT_GUIDES = {
    type: 'FeatureCollection',
    features: []
};
const DEFAULT_TOOLTIPS = [];
export class GeoJsonEditMode {
    _clickSequence = [];
    getGuides(props) {
        return DEFAULT_GUIDES;
    }
    getTooltips(props) {
        return DEFAULT_TOOLTIPS;
    }
    getSelectedFeature(props) {
        if (props.selectedIndexes.length === 1) {
            return props.data.features[props.selectedIndexes[0]];
        }
        return null;
    }
    getSelectedGeometry(props) {
        const feature = this.getSelectedFeature(props);
        if (feature) {
            return feature.geometry;
        }
        return null;
    }
    getSelectedFeaturesAsFeatureCollection(props) {
        const { features } = props.data;
        const selectedFeatures = props.selectedIndexes.map((selectedIndex) => features[selectedIndex]);
        return {
            type: 'FeatureCollection',
            features: selectedFeatures
        };
    }
    getClickSequence() {
        return this._clickSequence;
    }
    addClickSequence({ mapCoords }) {
        this._clickSequence.push(mapCoords);
    }
    resetClickSequence() {
        this._clickSequence = [];
    }
    getTentativeGuide(props) {
        const guides = this.getGuides(props);
        return guides.features.find((f) => f.properties && f.properties.guideType === 'tentative');
    }
    isSelectionPicked(picks, props) {
        if (!picks.length)
            return false;
        const pickedFeatures = getNonGuidePicks(picks).map(({ index }) => index);
        const pickedHandles = getPickedEditHandles(picks).map(({ properties }) => properties.featureIndex);
        const pickedIndexes = new Set([...pickedFeatures, ...pickedHandles]);
        return props.selectedIndexes.some((index) => pickedIndexes.has(index));
    }
    rewindPolygon(feature) {
        const { geometry } = feature;
        const isPolygonal = geometry.type === 'Polygon' || geometry.type === 'MultiPolygon';
        if (isPolygonal) {
            // @ts-expect-error turf type too wide
            return rewind(feature);
        }
        return feature;
    }
    getAddFeatureAction(featureOrGeometry, features) {
        // Unsure why flow can't deal with Geometry type, but there I fixed it
        const featureOrGeometryAsAny = featureOrGeometry;
        const feature = featureOrGeometryAsAny.type === 'Feature'
            ? featureOrGeometryAsAny
            : {
                type: 'Feature',
                properties: {},
                geometry: featureOrGeometryAsAny
            };
        const rewindFeature = this.rewindPolygon(feature);
        const updatedData = new ImmutableFeatureCollection(features)
            .addFeature(rewindFeature)
            .getObject();
        return {
            updatedData,
            editType: 'addFeature',
            editContext: {
                featureIndexes: [updatedData.features.length - 1]
            }
        };
    }
    getAddManyFeaturesAction({ features: featuresToAdd }, features) {
        let updatedData = new ImmutableFeatureCollection(features);
        const initialIndex = updatedData.getObject().features.length;
        const updatedIndexes = [];
        for (const feature of featuresToAdd) {
            const { properties, geometry } = feature;
            const geometryAsAny = geometry;
            updatedData = updatedData.addFeature({
                type: 'Feature',
                properties,
                geometry: geometryAsAny
            });
            updatedIndexes.push(initialIndex + updatedIndexes.length);
        }
        return {
            updatedData: updatedData.getObject(),
            editType: 'addFeature',
            editContext: {
                featureIndexes: updatedIndexes
            }
        };
    }
    getAddFeatureOrBooleanPolygonAction(featureOrGeometry, props) {
        const featureOrGeometryAsAny = featureOrGeometry;
        const selectedFeature = this.getSelectedFeature(props);
        const { modeConfig } = props;
        if (modeConfig && modeConfig.booleanOperation) {
            if (!selectedFeature ||
                (selectedFeature.geometry.type !== 'Polygon' &&
                    selectedFeature.geometry.type !== 'MultiPolygon')) {
                // eslint-disable-next-line no-console,no-undef
                console.warn('booleanOperation only supported for single Polygon or MultiPolygon selection');
                return null;
            }
            const feature = featureOrGeometryAsAny.type === 'Feature'
                ? featureOrGeometryAsAny
                : {
                    type: 'Feature',
                    geometry: featureOrGeometryAsAny
                };
            let updatedGeometry;
            if (modeConfig.booleanOperation === 'union') {
                // @ts-expect-error selectedFeature type too wide
                updatedGeometry = turfUnion(selectedFeature, feature);
            }
            else if (modeConfig.booleanOperation === 'difference') {
                // @ts-expect-error selectedFeature type too wide
                updatedGeometry = turfDifference(selectedFeature, feature);
            }
            else if (modeConfig.booleanOperation === 'intersection') {
                // @ts-expect-error selectedFeature type too wide
                updatedGeometry = turfIntersect(selectedFeature, feature);
            }
            else {
                // eslint-disable-next-line no-console,no-undef
                console.warn(`Invalid booleanOperation ${modeConfig.booleanOperation}`);
                return null;
            }
            if (!updatedGeometry) {
                // eslint-disable-next-line no-console,no-undef
                console.warn('Canceling edit. Boolean operation erased entire polygon.');
                return null;
            }
            const featureIndex = props.selectedIndexes[0];
            const updatedData = new ImmutableFeatureCollection(props.data)
                .replaceGeometry(featureIndex, updatedGeometry.geometry)
                .getObject();
            const editAction = {
                updatedData,
                editType: 'unionGeometry',
                editContext: {
                    featureIndexes: [featureIndex]
                }
            };
            return editAction;
        }
        return this.getAddFeatureAction(featureOrGeometry, props.data);
    }
    createTentativeFeature(props) {
        return null;
    }
    handleClick(event, props) { }
    handlePointerMove(event, props) {
        const tentativeFeature = this.createTentativeFeature(props);
        if (tentativeFeature) {
            props.onEdit({
                updatedData: props.data,
                editType: 'updateTentativeFeature',
                editContext: {
                    feature: tentativeFeature
                }
            });
        }
    }
    handleStartDragging(event, props) { }
    handleStopDragging(event, props) { }
    handleDragging(event, props) { }
    handleKeyUp(event, props) {
        if (event.key === 'Escape') {
            this.resetClickSequence();
            props.onEdit({
                // Because the new drawing feature is dropped, so the data will keep as the same.
                updatedData: props.data,
                editType: 'cancelFeature',
                editContext: {}
            });
        }
    }
}
export function getIntermediatePosition(position1, position2) {
    const intermediatePosition = [
        (position1[0] + position2[0]) / 2.0,
        (position1[1] + position2[1]) / 2.0
    ];
    return intermediatePosition;
}
