// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import { ArrowStyles, DEFAULT_STYLE, MAX_ARROWS } from "../style.js";
import { NebulaLayer } from "../nebula-layer.js";
import { toDeckColor } from "../../utils/utils.js";
import { DeckCache } from "../deck-renderer/deck-cache.js";
import { PathMarkerLayer } from '@deck.gl-community/layers';
const NEBULA_TO_DECK_DIRECTIONS = {
    [ArrowStyles.NONE]: { forward: false, backward: false },
    [ArrowStyles.FORWARD]: { forward: true, backward: false },
    [ArrowStyles.BACKWARD]: { forward: false, backward: true },
    [ArrowStyles.BOTH]: { forward: true, backward: true }
};
export class SegmentsLayer extends NebulaLayer {
    deckCache;
    noBlend;
    highlightColor;
    arrowSize;
    jointRounded;
    capRounded;
    dashed;
    markerLayerProps;
    constructor(config) {
        super(config);
        this.deckCache = new DeckCache(config.getData, (data) => config.toNebulaFeature(data));
        this.enableSelection = true;
        this.enablePicking = config.enablePicking ?? true;
        this.noBlend = config.noBlend ?? false;
        this.jointRounded = config.jointRounded ?? true;
        this.capRounded = config.capRounded ?? true;
        this.dashed = config.dashed ?? false;
        this.markerLayerProps = config.markerLayerProps ?? null;
        this.highlightColor = [1, 1, 1, 1];
        this.arrowSize = 1;
    }
    getMouseOverSegment() {
        // TODO: remove references
        return null;
    }
    _calcMarkerPercentages(nf) {
        const { arrowPercentages } = nf.style;
        if (arrowPercentages) {
            return arrowPercentages;
        }
        const arrowStyle = nf.style.arrowStyle || DEFAULT_STYLE.arrowStyle;
        if (arrowStyle === ArrowStyles.NONE)
            return [];
        const arrowCount = Math.min(nf.style.arrowCount || DEFAULT_STYLE.arrowCount, MAX_ARROWS);
        return [[0.5], [0.33, 0.66], [0.25, 0.5, 0.75]][arrowCount - 1];
    }
    _getHighlightedObjectIndex({ nebula }) {
        const { deckglMouseOverInfo } = nebula;
        if (deckglMouseOverInfo) {
            const { originalLayer, index } = deckglMouseOverInfo;
            if (originalLayer === this) {
                return index;
            }
        }
        // no object
        return -1;
    }
    render({ nebula }) {
        const defaultColor = [0x0, 0x0, 0x0, 0xff];
        const { objects, updateTrigger } = this.deckCache;
        return new PathMarkerLayer({
            id: `segments-${this.id}`,
            data: objects,
            opacity: 1,
            fp64: false,
            jointRounded: this.jointRounded,
            capRounded: this.capRounded,
            pickable: true,
            sizeScale: this.arrowSize || 6,
            parameters: {
                depthCompare: 'always',
                blendColorOperation: this.noBlend ? undefined : 'max'
            },
            getPath: (nf) => nf.geoJson.geometry.coordinates,
            getColor: (nf) => toDeckColor(nf.style.lineColor, defaultColor),
            getWidth: (nf) => nf.style.lineWidthMeters || 1,
            getZLevel: (nf) => nf.style.zLevel * 255,
            getDirection: (nf) => NEBULA_TO_DECK_DIRECTIONS[nf.style.arrowStyle],
            getMarkerColor: (nf) => toDeckColor(nf.style.arrowColor, defaultColor),
            // eslint-disable-next-line @typescript-eslint/unbound-method
            getMarkerPercentages: this._calcMarkerPercentages,
            updateTriggers: { all: updateTrigger },
            highlightedObjectIndex: this._getHighlightedObjectIndex({ nebula }),
            highlightColor: toDeckColor(this.highlightColor),
            dashJustified: this.dashed,
            getDashArray: this.dashed ? (nf) => nf.style.dashArray : undefined,
            markerLayerProps: this.markerLayerProps ||
                PathMarkerLayer.defaultProps.markerLayerProps,
            nebulaLayer: this
        });
    }
}
