// deck.gl-community
// SPDX-License-Identifier: MIT
// Copyright (c) vis.gl contributors
import { ModeHandler, getPickedEditHandle, getEditHandlesForGeometry } from "./mode-handler.js";
// TODO edit-modes: delete handlers once EditMode fully implemented
export class SnappableHandler extends ModeHandler {
    _handler;
    _editHandlePicks;
    _startDragSnapHandlePosition = undefined;
    constructor(handler) {
        super();
        this._handler = handler;
    }
    setFeatureCollection(featureCollection) {
        this._handler.setFeatureCollection(featureCollection);
    }
    setModeConfig(modeConfig) {
        this._modeConfig = modeConfig;
        this._handler.setModeConfig(modeConfig);
    }
    setSelectedFeatureIndexes(indexes) {
        this._handler.setSelectedFeatureIndexes(indexes);
    }
    _getSnappedMouseEvent(event, snapPoint) {
        // @ts-expect-error narrow event type
        return Object.assign({}, event, {
            mapCoords: snapPoint,
            pointerDownMapCoords: this._startDragSnapHandlePosition
        });
    }
    _getEditHandlePicks(event) {
        const { picks } = event;
        const potentialSnapHandle = picks.find((pick) => pick.object && pick.object.type === 'intermediate');
        const handles = { potentialSnapHandle: potentialSnapHandle && potentialSnapHandle.object };
        const pickedHandle = getPickedEditHandle(event.pointerDownPicks);
        if (pickedHandle) {
            return { ...handles, pickedHandle };
        }
        return handles;
    }
    _updatePickedHandlePosition(editAction) {
        const { pickedHandle = {} } = this._editHandlePicks || {};
        if (pickedHandle && editAction) {
            const { featureIndexes, updatedData } = editAction;
            for (let i = 0; i < featureIndexes.length; i++) {
                const selectedIndex = featureIndexes[i];
                const updatedFeature = updatedData.features[selectedIndex];
                const { positionIndexes, featureIndex } = pickedHandle;
                if (selectedIndex >= 0 && featureIndex === selectedIndex) {
                    const { coordinates } = updatedFeature.geometry;
                    pickedHandle.position = positionIndexes.reduce((a, b) => a[b], coordinates);
                }
            }
        }
    }
    // If additionalSnapTargets is present in modeConfig and is populated, this
    // method will return those features along with the features
    // that live in the current layer. Otherwise, this method will simply return the
    // features from the current layer
    _getSnapTargets() {
        let { additionalSnapTargets } = this.getModeConfig() || {};
        additionalSnapTargets = additionalSnapTargets || [];
        const features = [
            ...this._handler.featureCollection.getObject().features,
            ...additionalSnapTargets
        ];
        return features;
    }
    _getNonPickedIntermediateHandles() {
        const handles = [];
        const features = this._getSnapTargets();
        for (let i = 0; i < features.length; i++) {
            // Filter out the currently selected feature(s)
            const isCurrentIndexFeatureNotSelected = i < features.length && !this._handler.getSelectedFeatureIndexes().includes(i);
            if (isCurrentIndexFeatureNotSelected) {
                const { geometry } = features[i];
                handles.push(...getEditHandlesForGeometry(geometry, i, 'intermediate'));
            }
        }
        return handles;
    }
    // If no snap handle has been picked, only display the edit handles of the
    // selected feature. If a snap handle has been picked, display said snap handle
    // along with all snappable points on all non-selected features.
    getEditHandles(picks, mapCoords) {
        const { enableSnapping } = this._modeConfig || {};
        const handles = this._handler.getEditHandles(picks, mapCoords);
        if (!enableSnapping)
            return handles;
        const { pickedHandle } = this._editHandlePicks || {};
        if (pickedHandle) {
            handles.push(...this._getNonPickedIntermediateHandles(), pickedHandle);
            return handles;
        }
        const { features } = this._handler.featureCollection.getObject();
        for (const index of this._handler.getSelectedFeatureIndexes()) {
            if (index < features.length) {
                const { geometry } = features[index];
                handles.push(...getEditHandlesForGeometry(geometry, index, 'snap'));
            }
        }
        return handles.filter(Boolean);
    }
    _getSnapAwareEvent(event) {
        const { potentialSnapHandle } = this._editHandlePicks || {};
        return potentialSnapHandle && potentialSnapHandle.position
            ? this._getSnappedMouseEvent(event, potentialSnapHandle.position)
            : event;
    }
    handleStartDragging(event) {
        this._startDragSnapHandlePosition = (getPickedEditHandle(event.picks) || {}).position;
        return this._handler.handleStartDragging(event);
    }
    handleStopDragging(event) {
        // @ts-expect-error narrow event type
        const modeActionSummary = this._handler.handleStopDragging(this._getSnapAwareEvent(event));
        this._editHandlePicks = null;
        return modeActionSummary;
    }
    getCursor(event) {
        return this._handler.getCursor(event);
    }
    handlePointerMove(event) {
        const { enableSnapping } = this._handler.getModeConfig() || {};
        if (enableSnapping) {
            this._editHandlePicks = this._getEditHandlePicks(event);
        }
        // @ts-expect-error narrow event type
        const modeActionSummary = this._handler.handlePointerMove(this._getSnapAwareEvent(event));
        const { editAction } = modeActionSummary;
        if (editAction) {
            this._updatePickedHandlePosition(editAction);
        }
        return modeActionSummary;
    }
}
